% Haskell: From a parallel perspective
% Brian McKenna
% October 13, 2010

# What's Haskell?

- Functional
- Pure (referentially transparent)
- Lazy (non-strict)
- Strong static typing
    - More strong than Java or C#
- Open standard to consolidate other languages with above features
- Largest implementation is GHC
- Less important to this talk
    - Pattern matching, algebraic data types, type inference, currying

# What's it used for?

[
![Frag](images/Frag1.png)
][Frag] [
![xmonad](images/Screen-mauke-dzen-small.png)
][xmonad] [
![OmegaGB](images/omegagb-0005.png)
][OmegaGB] [
![House](images/screendump-oct2005-small.png)
][House] [
![Leksah](images/osx-leksah-0.8-small.png)
][Leksah] [
![Blobs](images/dazzle.jpg)
][Blobs] [
![Hircules](images/Screenshot-Hircules.png)
][Hircules]

# How's it look?

~~~ {.haskell}
main = putStrLn "Hello world"
~~~

# How's it look?

~~~ {.haskell}
fac 0 = 1
fac n = n * fac (n-1)

main = print (fac 42)
~~~

# Why parallel Haskell?

- Not sequential by default
    - Sequence is usually through function application
- Can be deterministic
    - Dependencies and side effects are explicit
    - (Data is immutable and functions return side effects)
- Many different methods of parallelism
- *Not magic*
    - Sequential algorithms still have sequential dependencies

# Explicit side effects

I really want to stress this point:

~~~ {.haskell}
f :: Int -> Int
g :: Int -> IO Int
~~~

Sequence is just function application. The following are the same:

~~~ {.haskell}
main = do
  name <- getLine
  putStrLn ("Hi " ++ name)

main = getLine >>=
       (\name -> putStrLn ("Hi " ++ name))
~~~

# Popular parallelism

## Explicit

- Threads
- Locks

## Implicit

- Some languages have built-in parallel constructs
- Data parallelism

## Semi-implicit

- Say *when* to run in parallel but not *how*

# Explicit concurrency (Java)

~~~ {.java}
import java.lang.Runnable;

class ThreadTest implements Runnable {
    public void run() {
        System.out.println("Hello world");
    }

    public static void main(String[] args) {
        Thread runner = new Thread(new ThreadTest());
        runner.start();
        try {
            runner.join();
        } catch(InterruptedException e) {
        }
    }
}
~~~

# Explicit concurrency (Clojure)

~~~ {.scheme}
(let [runner (Thread. #(println "Hello world"))]
  (.start runner)
  (.join runner))
~~~

# Semi-implicit parallelism (C# - TPL)

~~~ {.java}
using System;
using System.Threading.Tasks;

class Test
{
    static void Main()
    {
        Parallel.For(0, 100, i =>
        {
            Console.WriteLine(i);
        });
    }
}
~~~

# Implicit parallelism (NESL - nested data )

~~~ {.scheme}
function qsort(a) =
if (#a < 2) then a
else
  let pivot   = a[#a/2];
      lesser  = {e in a| e < pivot};
      equal   = {e in a| e == pivot};
      greater = {e in a| e > pivot};
      result  = {qsort(v): v in [lesser,greater]};
  in result[0] ++ equal ++ result[1] $
~~~

# Haskell is diverse

## Semi-implicit (Parallel Haskell)

- `par` and `pseq` combinators
- Strategies and skeletons (`parList`, `parMap`, `parZipWith`, etc)

## Implicit (Data Parallel Haskell)

- Flat/nested data parallelism
- Intel CnC

## Explicit concurrency (Concurrent Haskell)

- `forkIO` with `MVar`, `Chan`, `QSemN` and `STM`

# Semi-implicit in Haskell

# Quicksort (sequential)

~~~ {.haskell}


qsort [] = []
qsort [x] = [x]
qsort (x:xs) = lt ++ (x:gt)
    where
      lt = qsort [y | y <- xs, y < x]
      gt = qsort [y | y <- xs, y >= x]

main = do
  print $ head sorted -- Will be 1
  print $ last sorted -- Will be 10000
    where nums = [1..5000] ++ reverse [5000..10000]
          sorted = qsort nums
~~~

# Quicksort (with `par` and `pseq`)

~~~ {.haskell}
import Control.Parallel

qsort [] = []
qsort [x] = [x]
qsort (x:xs) = lt `par` gt `pseq` lt ++ (x:gt)
    where
      lt = qsort [y | y <- xs, y < x]
      gt = qsort [y | y <- xs, y >= x]

main = do
  print $ head sorted -- Will be 1
  print $ last sorted -- Will be 10000
    where nums = [1..5000] ++ reverse [5000..10000]
          sorted = qsort nums
~~~

# `par` and `pseq`

## `par a b`

- Creates a "spark" to evaluate `a` (non-lazily/strictly)
    - Sparks are put in a queue to be run on an OS thread
- Returns `b` immediately

## `pseq a b`

- Non-lazily/strictly evaluates `a` before it starts to evaluate `b`

~~~ {.haskell}
a `par` b `pseq` a + b
~~~

- During evaluation of `b`, the `a` spark can be scheduled on a separate thread
- After `b` is finished, `a + b` will just use the already computed values

# Algorithm + Strategy = Parallelism

## Skeletons

- `parList` - evaluates each element in a list
- `parListChunk` - splits a list equally and evaluates each chunk
- `parMap` - evaluates a function over each element in a list
- And many more...

## Evaluation degrees

- `r0` - don't evaluate
- `rwhnf` - evaluate to weak head normal form
- `rnf` - evaluate to head normal form

# `par`, `pseq` and stategies

## Why?

- Easy (simple combinators)
- Cheap (sparks put in queue)
- Deterministic (only changes evaluation)
    - Worst case, it runs slower

### Careful!

- Sparks are cheap, not free.
    - Only create sparks when cost is much less than evaluation
    - Theshold is common tactic for divide and conquer

# Implicit in Haskell

# Remember the NESL example?

~~~ {.scheme}
function qsort(a) =
if (#a < 2) then a
else
  let pivot   = a[#a/2];
      lesser  = {e in a| e < pivot};
      equal   = {e in a| e == pivot};
      greater = {e in a| e > pivot};
      result  = {qsort(v): v in [lesser,greater]};
  in result[0] ++ equal ++ result[1] $
~~~

# Quicksort (Data Parallel Haskell)

~~~ {.haskell}
{-# LANGUAGE PArr #-}
{-# OPTIONS -fvectorise #-}
{-# OPTIONS -fno-spec-constr-count #-}
module QSort (qsort) where

import Data.Array.Parallel.Prelude
import Data.Array.Parallel.Prelude.Int

import qualified Prelude

qsort :: PArray Int -> PArray Int
{-# NOINLINE qsort #-}
qsort xs = toPArrayP (qsort' (fromPArrayP xs))
~~~

# Quicksort (Data Parallel Haskell)

~~~ {.haskell}
qsort' :: [:Int:] -> [:Int:]
qsort' a
    | lengthP a < 2 = a
    | otherwise =
        let pivot   = a !: (lengthP a `div` 2)
            lesser  = [:e | e <- a, e < pivot:]
            equal   = [:e | e <- a, e == pivot:]
            greater = [:e | e <- a, e > pivot:]
            result  = mapP qsort' [:lesser, greater:]
        in (result !: 0) +:+ equal +:+ (result !: 1)
~~~

# Data Parallel Haskell

- Experimental
- Highly influenced by NESL
- Only possible in purely functional languages
- Distributed memory eventually
- GPU possibly in the future
- Not impressively fast

# Intel Concurrent Collections (CnC)

~~~ {.haskell}
myStep items tag =
  do word1 <- get items "left"
     word2 <- get items "right"
     put items "result"
         (word1 ++ word2 ++ show tag)

cncGraph =
  do tags  <- newTagCol
     items <- newItemCol
     prescribe tags (myStep items)
     initialize$
        do put items "left"  "Hello "
	   put items "right" "World "
	   putt tags 99
     finalize$
        do get items "result"
~~~

# Intel Concurrent Collections (CnC)

- Coordination language
    - Data flow
    - Control flow
- Many different schedulers provided
    - 4 to choose from
    - Provide same API, only different backends
- Similar to C++ CnC but benefits from Haskell's unique features
- Influenced by SISAL
- Scales well

# Explicit in Haskell

# Concurrent Haskell

- `forkIO` - thread spawning
- `MVar` - mutable values
    - `SampleVar` is a mutable variable that overwrites data
- `Chan` - unbounded channels
- `QSemN` - semaphores
    - `QSem` is a semaphore of one (mutex)

# Creating a thread (`forkIO`)

~~~ {.haskell}
import Control.Concurrent

main = do
  putStrLn "Hello"
  forkIO (putStrLn "world")
  threadDelay 1000000 -- Let the thread finish
~~~

# Mutable values (`MVar`)

~~~ {.haskell}
import Control.Concurrent
import Control.Concurrent.MVar

child mvar = do
  putStrLn "world"
  threadDelay 1000000
  putMVar mvar ()

main = do
  putStrLn "Hello"
  mvar <- newEmptyMVar
  forkIO (child mvar)
  takeMVar mvar
  return ()
~~~

# Channels (`Chan`)

~~~ {.haskell}
import Control.Concurrent
import Control.Concurrent.Chan

main = do
  chan <- newChan
  forkIO (child chan)
  threadDelay 1000000
  printNumbers chan
  return ()
~~~

# Channels (`Chan`)

~~~ {.haskell}
child chan = do
  mapM_ (writeChan chan) [2,4..100]
  return ()

printNumbers chan = do
  empty <- isEmptyChan chan
  if empty
     then return ()
     else do
        n <- readChan chan
        print n
        threadDelay 1000000
        printNumbers chan
~~~

# Semaphores (`QSemN`)

~~~ {.haskell}
import Control.Concurrent
import Control.Concurrent.QSemN

child sem n = do
  threadDelay (1000000 * n)
  putStrLn ("world " ++ show n)
  signalQSemN sem 1

main = do
  sem <- newQSemN 0
  putStrLn "Hello"
  mapM_ (\n -> forkIO (child sem n)) [1..4]
  waitQSemN sem 4
~~~

# Software Transactional Memory (STM)

- Composable
    - Nested actions act as one
- Modular
- Provide the ACI from ACID
    - Atomicity
    - Consistency
    - Isolation
    - (There is no durability because it's in memory)
- Lock safe

# STM variables (`TVar`)

~~~ {.haskell}
import Control.Concurrent
import Control.Concurrent.STM

main = do
  var <- atomically $ newTVar 0
  mapM_ (\_ -> forkIO $ incValue var 10000) [1..4]
  printWhileLess var 40000
~~~  

# STM variables (`TVar`)

~~~ {.haskell}
incValue var times = do
  atomically $ do
    n <- readTVar var
    writeTVar var (n + 1)
  if times == 0
     then return ()
     else incValue var (times - 1)

printWhileLess var times = do
  n <- atomically $ readTVar var
  if n >= times
     then return ()
     else do
       print n
       printWhileLess var times
~~~

# More STM

## Other functions

- `retry` - forces a rollback and retry of the transaction
- `check a` - checks boolean `a` and retries on false
- `orElse a b` - tries transaction `a` and on failure, tries transaction `b`

## Other data types

- `TMVar` - transactional `MVar`
- `TChan` - transactional `Chan`
- `TArray` - transactional `MArray`

# ThreadScope

![](images/ThreadScope.png)

# Summary

- Haskell encourages parallelism
- Many different options exist
    - Implicit, semi-implicit, explicit
    - Each with their own benefits
    - Everything I showed today only works on shared memory mulitprocessors
- You can get very far without using locks
- Threading is very lightweight (better than OS threads)
    - Sparks are even cheaper
- Nice tools hooking into the GHC runtime for profiling parallelism

# Thanks

## Brian McKenna

[http://brianmckenna.org/][Site]

[http://twitter.com/puffnfresh][Twitter]

## Special thanks to

- Miran Lipovača ([@bonus500][bonus500])
    - For his beautiful illustrations from [LYAH][]

# References

- Why Functional Programming Matters - _John Hughes_
- [Learn You a Haskell for Great Good!][LYAH] - _Miran Lipovača_
- Seq no more: Better Strategies for Parallel Haskell - _Simon Marlow, Patrick Maier, Hans-Wolfgang Loidl, Mustafa Aswad and Phil Trinder_
- Accidents always Come in Threes: A Case Study of Data-intensive Programs in Parallel Haskell - _Philip W. Trinder, Kevin Hammond, Hans-Wolfgang Loidl, Simon L. Peyton Jones and J. Wu_
- Algorithm + Strategy = Parallelism - _Philip W. Trinder, Kevin Hammond, Hans-Wolfgang Loidl and Simon L. Peyton Jones_
- Harnessing the Multicores: Nested Data Parallelism in Haskell - _Simon Peyton Jones, Roman Leshchinskiy, Gabriele Keller, Manuel M. T. Chakravarty_

# References

- Intel Concurrent Collections for Haskell - _Ryan Newton, Chih-Ping Chen and Simon Marlow_
- A Tutorial on Parallel and Concurrent Programming in Haskell - _Simon Peyton Jones and Satnam Singh_
- Beautiful concurrency - _Simon Peyton Jones_
- Composable Memory Transactions - _Tim Harris, Simon Marlow, Simon Peyton Jones and Maurice Herlihy_
- Parallel Performance Tuning for Haskell - _Don Jones Jr., Simon Marlow and Satnam Singh_

[Frag]: http://www.haskell.org/haskellwiki/Frag
[OmegaGB]: http://www.mutantlemon.com/omegagb/
[xmonad]: http://xmonad.org/
[House]: http://programatica.cs.pdx.edu/House/
[Leksah]: http://leksah.org/
[Blobs]: http://www.cs.york.ac.uk/fp/darcs/Blobs/
[Hircules]: http://www.haskell.org/hircules/
[bonus500]: http://twitter.com/bonus500
[LYAH]: http://learnyouahaskell.com/
[Site]: http://brianmckenna.org/
[Twitter]: http://twitter.com/puffnfresh
